import base64
import email.utils
import hashlib
import hmac
import logging
import requests

from . import config

log = logging.getLogger(__name__)


class Client:
    def __init__(self, endpoint, access_key, secret_key):
        self.endpoint = endpoint
        self.access_key = access_key
        self.secret_key = secret_key

    def __auth(self, method, date, headers, access_key, secret_key, uri):
        def getJssHeaders(headers):
            map = []
            for key, value in headers.items():
                if key.startswith(config.JSS_CONFIG['JSS_HEADER_PREFIX']):
                    map.append({key: value})
            return map

        headers_content_md5 = ''
        headers_content_type = headers.get('content-type', '')
        str_to_sign = method + '\n' + headers_content_md5 + '\n' + headers_content_type + '\n' + \
                      headers['date'] + '\n' + uri
        secret_key_encode = secret_key.encode('utf-8')
        h = hmac.new(secret_key_encode, str_to_sign.encode('utf-8'), hashlib.sha1)
        token = base64.b64encode(h.digest())
        return 'jingdong ' + access_key + ':' + token.decode()

    def execute(self, opt):
        date = email.utils.formatdate(usegmt=True)
        request_headers = {'date': date}
        # if additional parameters in headers exist, combine headers information together
        if 'headers' in opt:
            request_headers = {**request_headers, **opt['headers']}

        authorization = self.__auth(opt['method'], date, request_headers, self.access_key, self.secret_key, opt['uri'])
        request_headers['authorization'] = authorization

        log.debug("request header is: ---> {0}".format(request_headers))

        if opt['method'] == "GET":
            url = self.endpoint + opt['uri']
            log.debug("composed url is: " + url)
            if 'json' in opt and 'additional_params' in opt:
                return requests.get(url, params=opt['additional_params'], headers=request_headers)
            else:
                return requests.get(url, headers=request_headers)

        elif opt['method'] == "PUT":
            url = self.endpoint + opt['uri']
            log.debug("composed url is: " + url)
            return requests.put(url, headers=request_headers, data=opt['body'])

        elif opt['method'] == "DELETE":
            url = self.endpoint + opt['uri']
            log.debug("composed url is: " + url)
            return requests.delete(url, headers=request_headers)

        elif opt['method'] == "HEAD":
            url = self.endpoint + opt['uri']
            log.debug("composed url is: " + url)
            return requests.head(url, headers=request_headers)

        elif opt['method'] == "POST":
            url = self.endpoint + opt['uri']
            log.debug("composed url is: " + url)
            if 'body' in opt:
                return requests.post(url, headers=request_headers, data=opt['body'])
            else:
                return requests.post(url, headers=request_headers)
        else:
            raise ValueError("this method: " + opt['method'] + " is not valid")



import os
import urllib.parse
from requests import codes
from hashlib import md5
import logging
import json
import io

log = logging.getLogger(__name__)


class Object:
    """
    Object 是京东云存储中的基本实体，由键(Key)，数据(Data)和元数据(Metadata)三部分组成。关于数据(Data)，
    京东云存储并不关心其内容具体是什么。而元数据(Metadata)则是一组键值(Key-Value)的组合，包括数据的长度，创建时间，MD5 值等
    """

    def __init__(self, client, bucket, key):
        """创建一个 Object 实例
             @public
             :param {Client} client Client 实例
             :param {Bucket} bucket Bucket 实例
             :param {string} key Bucket 名称
        """

        if not (isinstance(key, str) and key.strip()):
            raise ValueError('name cannot be empty')
        self.key = key
        self.client = client
        self.bucket = bucket

    def get(self, headers=None):
        """下载一个资源并读取其内容
             @public
             :returns {Response.Body} 成功时 statusCode 为 200, 获得文件内容
             @example
             jss.bucket('your-bucket-name').object('your-object-name').get()
        """
        bucket_name = self.bucket.name
        object_key = self.key

        request = {
            'method': 'GET',
            'uri': '/' + bucket_name + '/' + object_key
        }
        if headers:
            request['headers'] = headers
        try:
            return self.client.execute(request)
        except Exception as e:
            logging.error("JSS: get object failed" + repr(e), stack_info=True)
            raise ConnectionError('JSS: get object failed!' + repr(e))

    def put_data(self, body_data, file_new_name):
        """以data方式上传或替换一个资源
             @public
             :param {string} data_body 要上传资源的raw data
             :param {string} file_new_name 下载时文件的保存名称
             :returns 成功时 statusCode 为 200, 返回的Body为空;
             @example
             jss.bucket('pro-test').object('123.html').put(your_data_variable, '123.html')
        """
        bucket_name = self.bucket.name
        object_key = file_new_name
        headers = {
            'content-type': '',
            'Content-Disposition': "attachment; filename={func_result}".format(
                func_result=urllib.parse.quote(file_new_name, safe='~()*!.\''))
        }
        request = {
            'method': 'PUT',
            'uri': '/' + bucket_name + '/' + object_key,
            'headers': headers,
            'body': body_data.encode(encoding='UTF-8')
        }
        try:
            return self.client.execute(request)
        except Exception as e:
            logging.error("JSS: put_data failed" + repr(e), stack_info=True)
            raise ConnectionError("put data failed" + repr(e))

    def put_file(self, file_with_path, file_new_name):
        """以本地文件形式上传或替换一个资源
             @public
             :param {string} file_with_path 要上传资源的本地路径
             :param {string} file_new_name 下载时文件的保存名称
             :returns {Promise.<Response.body>} resolve，成功时 statusCode 为 200, 返回的Body为空;
             @example
             jss.bucket('pro-test').object('123.html').put_file('/file/folder/123.html', '123.html')
        """
        bucket_name = self.bucket.name
        object_key = file_new_name

        if not (isinstance(file_with_path, str) and file_with_path.strip()):
            raise ValueError('参数 path 必须为非空字符串')

        if file_new_name is None:
            file_new_name = object_key

        if not os.path.exists(file_with_path):
            raise IOError('can not find file')

        headers = {
            'content-type': '',
            'Content-Disposition': "attachment; filename={func_result}".format(
                func_result=urllib.parse.quote(file_new_name, safe='~()*!.\''))
        }
        with open(file_with_path) as fh:
            data_body = fh.read()

            request = {
                'method': 'PUT',
                'uri': '/' + bucket_name + '/' + object_key,
                'headers': headers,
                'body': data_body.encode(encoding='UTF-8')
            }
            try:
                return self.client.execute(request)
            except Exception as e:
                logging.error("JSS: put_file failed" + repr(e), stack_info=True)
                raise ConnectionError("put file failed" + repr(e))

    def init_multipart_upload(self):
        """初始化分块上传任务
             @public
             :returns {string} upload_id 返回分块上传的任务id
        """
        bucket_name = self.bucket.name
        object_key = self.key

        request = {
            'method': 'POST',
            'uri': '/' + bucket_name + '/' + object_key + '?uploads'
        }
        res = self.client.execute(request)
        res.raise_for_status()
        json_text = json.loads(res.text)
        return json_text['UploadId']

    def upload_part(self, upload_id, part_number, input_stream):
        """上传分块数据
             @public
             :param {string} upload_id: 分块上传任务id
             :param {int} part_number: 自定义分块编号, e.g., 1, 2, 3...
             :returns {dict} headers 返回该分块的头文件信息，其中etag等信息需要用在complete_multipart_upload
        """
        if not (isinstance(upload_id, str) and upload_id.strip()):
            raise ValueError('upload_id 必须是非空字符串')
        if not (isinstance(part_number, int)):
            raise ValueError('part_number 必须是非空整数')
        bucket_name = self.bucket.name
        object_key = self.key

        request = {
            'method': 'PUT',
            'uri': '/' + bucket_name + '/' + object_key + '?partNumber=' + str(part_number) + '&uploadId=' + upload_id,
            'body': input_stream
        }
        res = self.client.execute(request)
        res.raise_for_status()
        return res.headers

    def complete_multipart_upload(self, upload_id, parts):
        """完成分块上传任务
             @public
             :param {string} upload_id: 分块上传任务id
             :param {List} parts: 分块信息，例子:
             parts = [{'PartNumber' : '1',
                        'ETag': "xxxxxxx"}]
             :returns {dict} 成功则返回Object的信息
        """
        if not (isinstance(upload_id, str) and upload_id.strip()):
            raise ValueError('upload_id 必须是非空字符串')

        bucket_name = self.bucket.name
        object_key = self.key

        headers = {
            'content-type': 'application/json'
        }
        part_body = {'Part': parts}
        request = {
            'method': 'POST',
            'uri': '/' + bucket_name + '/' + object_key + '?uploadId=' + upload_id,
            'headers': headers,
            'body': json.dumps(part_body)
        }
        res = self.client.execute(request)
        res.raise_for_status()
        json_text = json.loads(res.text)
        return json_text

    def abort_multipart_upload(self, upload_id):
        """终止分块上传任务
             @public
             :param {string} upload_id: initMultipartUpload() 成功后response body中的分块上传任务id
             :returns 成功时 statusCode 为 204, body为空
        """
        if not (isinstance(upload_id, str) and upload_id.strip()):
            raise ValueError('upload_id 必须是非空字符串')
        bucket_name = self.bucket.name
        object_key = self.key

        request = {
            'method': 'DELETE',
            'uri': '/' + bucket_name + '/' + object_key + '?uploadId=' + upload_id
        }
        try:
            return self.client.execute(request)
        except Exception as e:
            raise ConnectionError('abort Multi-part upload failed: ' + repr(e))

    def delete(self):
        """删除一个 Object
             @public
             :returns 成功时 statusCode 为 204, 返回的Body为空
        """

        bucket_name = self.bucket.name
        object_key = self.key
        request = {
            'method': 'DELETE',
            'uri': '/' + bucket_name + '/' + object_key
        }
        try:
            return self.client.execute(request)
        except:
            raise ValueError('delete bucket failed!')

    def head(self):
        """获取一个 Object的元数据
             @public
             :returns {dict} 返回Object的元数据
             @example
             jss.bucket('pro-test').object('123.html').head()
        """
        bucket_name = self.bucket.name
        object_key = self.key
        request = {
            'method': 'HEAD',
            'uri': '/' + bucket_name + '/' + object_key
        }
        res = self.client.execute(request)
        res.raise_for_status()
        return res.headers

    def exist(self):
        """判断某个 Object的是否存在
             @public
             :returns 成功时为True, 失败返回false
             @example
             jss.bucket('pro-test').object('123.html').head()
        """
        bucket_name = self.bucket.name
        object_key = self.key
        request = {
            'method': 'HEAD',
            'uri': '/' + bucket_name + '/' + object_key
        }

        r = self.client.execute(request)

        if r.status_code == codes.ok:
            return True
        else:
            return False



JSS_CONFIG = {
    'end_point': 'http://storage.jd.com',
    'access_key': '',
    'secret_key': '',
    'JSS_HEADER_PREFIX': 'x-jss-',
    'COPY_SOURCE': 'x-jss-copy-source',
    'X_JSS_REQUETS_ID': 'x-jss-request-id',
    'JSON_TYPE': 'application/json, charset=UTF-8',
    'X_JSS_OBJECT_EXPIRES': 'x-jss-object-expires',
    'X_JSS_COPY_SOURCE': 'x-jss-copy-source',
    'X_JSS_METADATA_DIRECTIVE': 'x-jss-metadata-directive',
    'X_JSS_COPY_SOURCE_IF_MATCH': 'x-jss-copy-source-if-match',
    'X_JSS_COPY_SOURCE_IF_NONE_MATCH': 'x-jss-copy-source-if-none-match',
    'X_JSS_COPY_SOURCE_IF_UNMODIFIED_SINCE': 'x-jss-copy-source-if-unmodified-since',
    'X_JSS_COPY_SOURCE_IF_MODIFIED_SINCE': 'x-jss-copy-source-if-modified-since',
    'X_JSS_COPY_RANGE': 'x-jss-copy-range'
}



import hashlib

def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()

def parse_jss_uri(uri):
    """Return the bucket and object path of the uri.

    Parameters:
        uri : string
            A JSS URI such as `jss://domainname/mybucket/path/to/file`

    Examples:
        >>> parse_jss_uri("jss://domainname/mybucket/path/to/file")
        'mybucket', 'path/to/file'
    """
    if uri.startswith('jss://'):
        uri = uri[6:]
    if '/' not in uri:
        return uri, ""
    elif uri.startswith('/'):
        uri = uri[1:]

    s = uri.split('/', 2)
    return s[1], s[2]



from .client import Client
from .bucket import Bucket
from .config import JSS_CONFIG
import json
import hashlib
import re


class Jss:

    def __init__(self, end_point, access_key, secret_key):
        """创建一个 jss 客户端实例
             @public
             :param {string} endpoint 所要连接的 endpoint 地址
             :param {string} accessKey 字符串 accessKey
             :param {string} secretKey 字符串 secretKey
             @example
             jss=new jss('storage.jd.com','your-accessKey-here','your-secretKey-here');
        """

        if not (isinstance(end_point, str) and end_point.strip()):
            if 'end_point' not in JSS_CONFIG or len(JSS_CONFIG['end_point']) == 0:
                raise ValueError('endpoint cannot be empty')
            else:
                end_point = JSS_CONFIG['end_point']
        if not (isinstance(access_key, str) and access_key.strip()):
            if 'access_key' not in JSS_CONFIG or len(JSS_CONFIG['access_key']) == 0:
                raise ValueError('access key cannot be empty')
            else:
                access_key = JSS_CONFIG['access_key']
        if not (isinstance(secret_key, str) and end_point.strip()):
            if 'secret_key' not in JSS_CONFIG or len(JSS_CONFIG['secret_key']) == 0:
                raise ValueError('secret key cannot be empty')
            else:
                secret_key = JSS_CONFIG['secret_key']

        self.endpoint = end_point
        self.access_key = access_key
        self.secret_key = secret_key
        self.client = Client(end_point, access_key, secret_key)

    def bucket(self, name):
        """新建一个 Bucket 对象, 该对象封装了对 Bucket 的操作
             @public
             :param {string} bucketName bucket的名字;
             :returns {Bucket} 新建的 Bucket 对象
             @example
             bucket=jss.bucket('bucket-name')
        """

        return Bucket(self.client, name)

    def list_buckets(self):
        """列出所有 bucket
             @public
             :returns [{Bucket}]
             @example
             bucket=jss.list_bucket()
        """

        request = {
            'method': 'GET',
            'uri': '/'
        }
        res = self.client.execute(request)
        res.raise_for_status()
        json_text = json.loads(res.text)
        return json_text['Buckets']

    def has_bucket(self, bucket_name):
        """ 判断bucket是否存在
             @public
             :param {string} bucket_name bucket的名字;
             :returns [{Bucket}]
             @example
             bucket=jss.has_bucket('your_bucket_name')
        """
        if not (isinstance(bucket_name, str) and bucket_name.strip()):
            raise ValueError('参数 name 必须为非空字符串')

        bucket_list = self.list_buckets()
        has_bucket_flag = False
        for item in bucket_list:
            if item['Name'] == bucket_name:
                has_bucket_flag = True
                break
        return has_bucket_flag

    def _encode_post_data(self, js_data):
        encode_data = []
        if not isinstance(js_data, dict):
            encode_data.extend([self._encode_post_data(i) for i in js_data])
            return encode_data
        for values in js_data.values():
            if isinstance(values, dict):
                encode_data.extend(self._encode_post_data(values))
            elif isinstance(values, list):
                for i in values:
                    if isinstance(i, dict):
                        encode_data.extend(self._encode_post_data(i))
                    else:
                        i = re.sub(u"([^\u0041-\u005a\u0061-\u007a\u0030-\u0039\u002e])", "", str(i))
                        if i == "":
                            continue
                        encode_data.append(i)
            else:
                values = re.sub(u"([^\u0041-\u005a\u0061-\u007a\u0030-\u0039\u002e])", "", str(values))
                if values == "":
                    continue
                encode_data.append(values)
        return encode_data    

    def save_data(self, url, post_data, resp, bucket_name="pfinder-data-bucket"):
        if post_data is None:
            file_name = '/'.join(url.split("/")[-2:])
        else:
            file_name = url.split("/")[-1] + '/' + '/'.join(self._encode_post_data(post_data))
        resp = self.bucket(bucket_name).object(file_name).put_data(json.dumps(resp), file_name)
        if resp.status_code != 200:
            raise OverflowError("Save data into oss error!")
        return True

    def get_data(self, url, post_data, bucket_name="pfinder-data-bucket"):
        if post_data is None:
            file_name = '/'.join(url.split("/")[-2:])
        else:
            file_name = url.split("/")[-1] + '/' + '/'.join(self._encode_post_data(post_data))
        resp = self.bucket(bucket_name).object(file_name).get().content
        data = json.loads(resp)
        return data

    def has_data(self, url, post_data, bucket_name="pfinder-data-bucket"):
        if post_data is None:
            file_name = '/'.join(url.split("/")[-2:])
        else:
            file_name = url.split("/")[-1] + '/' + '/'.join(self._encode_post_data(post_data))
        has_file_flag = self.bucket(bucket_name).object(file_name).exist()
        return has_file_flag



#!/usr/bin/env python 
# -*- coding: utf-8 -*-

"""
@Author  : ZuoXiang
@Email   : zx_data@126.com
@Time    : 2021/12/30 5:47 下午
@File    : setup.py
@Desc    : 
"""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jss_upload",
    version="0.1.3",
    author="zuoxiang",
    author_email="zx_data@126.com",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zuoxiang95/jss_upload",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from .object import Object
import json

"""
Bucket是存放“文件”（云存储称之为Object）的容器，所有的Object都必须存放到某个Bucket中。
"""


class Bucket:

    def __init__(self, client, name):
        """创建一个 Bucket 实例
             @public
             :param {Client} client Client 实例
             :param {string} name Bucket 名称
        """
        if not (isinstance(name, str) and name.strip()):
            raise ValueError('name cannot be empty')
        self.name = name
        self.client = client

    def create(self):
        """新建一个 Bucket
             @public
             :returns 成功时 statusCode 为 200, 返回的Body为空
        """
        request = {
            'method': 'PUT',
            'uri': '/' + self.name
        }
        try:
            return self.client.execute(request)
        except Exception:
            raise ValueError('cannot create bucket')

    def delete(self):
        """删除一个 Bucket
             @public
             :returns 成功时 statusCode 为 204, 返回的Body为空
        """
        request = {
            'method': 'DELETE',
            'uri': '/' + self.name
        }
        try:
            delete_result = self.client.execute(request)
            return delete_result
        except:
            raise ValueError('cannot delete bucket')

    def object(self, key):
        """创建一个 JssObject
             @public
             :param {string} key Object的名字;
             :returns {JssObject} 新建的 JssObject 对象
             @example
             object=jss.bucket('bucket-name').object('object-name')
        """
        return Object(self.client, self, key)

    def list_object(self, marker=None, maxKeys=None, prefix=None, delimiter=None):
        """列出所有 JssObject （可加筛选条件）
             @public
             :optional param {string} marker 指定的Object的Key的起始标志;
             :optional param {string} maxKeys 指定的Object的数量;
             :optional param {string} prefix 指定的Object Key的前缀;
             :optional param {string} delimiter 指定的Delimiter分组符;
             :returns [{JssObject}] 新建的 JssObject 对象
             @example
             object=jss.bucket('bucket-name').object('object-name')
        """
        name = self.name
        if (not marker and not maxKeys and not prefix and not delimiter):
            request = {
                'method': 'GET',
                'uri': '/' + name,
                'json': True
            }
        else:
            additional_params = {}
            if marker:
                additional_params['marker'] = marker
            if maxKeys:
                additional_params['maxKeys'] = maxKeys
            if prefix:
                additional_params['prefix'] = prefix
            if delimiter:
                additional_params['delimiter'] = delimiter
            request = {
                'method': 'GET',
                'uri': '/' + name,
                'json': True,
                'additional_params': additional_params
            }
        res = self.client.execute(request)
        res.raise_for_status()
        json_text = json.loads(res.text)
        obj_list, hasnext = json_text['Contents'], json_text['HasNext']
        while hasnext:
            request['additional_params']['marker'] = obj_list[-1]['Key']
            res = self.client.execute(request)
            res.raise_for_status()
            json_text = json.loads(res.text)
            obj_list, hasnext = obj_list + json_text['Contents'], json_text['HasNext']
        return obj_list

    # TODO: ACL
    # def acl(self):
